<?php
/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

namespace FiloBlu\OrderAdditionalData\Model\Plugin;

use FiloBlu\OrderAdditionalData\Model\AdditionalDataFactory;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Customer\Model\AddressFactory;
use Magento\Customer\Model\CustomerFactory;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\DataObject;
use Magento\Framework\DataObjectFactory;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Quote\Model\Quote\Address\RateRequestFactory;
use Magento\Sales\Api\Data\OrderExtensionFactory;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\OrderItemExtensionFactory;
use Magento\Sales\Api\OrderManagementInterface;
use Magento\Store\Model\ScopeInterface;
use WebShopApps\MatrixRate\Model\Carrier\MatrixrateFactory;

/**
 * Plugin that extends OrderManagement. Save Additional Data on order place.
 */
class OrderPlace
{

    /**
     * @var CustomerFactory
     */
    protected $customerFactory;

    /**
     * @var AddressFactory
     */
    protected $addressFactory;

    /**
     * @var  ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var RateRequestFactory
     */
    protected $rateRequestFactory;

    /**
     * @var MatrixrateFactory
     */
    protected $carrierFactory;

    /**
     * @var DataObjectFactory
     */
    protected $objectFactory;

    /**
     * @var OrderExtensionFactory
     */
    protected $orderExtensionFactory;

    /**
     * @var OrderItemExtensionFactory
     */
    protected $orderItemExtensionFactory;

    /**
     * @var SerializerInterface
     */
    protected $_serializer;

    /**
     * @var AdditionalDataFactory
     */
    protected $_additionalDataFactory;

    /**
     * OrderPlace constructor.
     * @param CustomerFactory $customerFactory
     * @param AddressFactory $addressFactory
     * @param ProductRepositoryInterface $productRepository
     * @param ScopeConfigInterface $scopeConfig
     * @param RateRequestFactory $rateRequestFactory
     * @param MatrixrateFactory $carrierFactory
     * @param DataObjectFactory $objectFactory
     * @param OrderExtensionFactory $orderExtensionFactory
     * @param OrderItemExtensionFactory $orderItemExtensionFactory
     * @param SerializerInterface $serializer
     * @param AdditionalDataFactory $additionalDataFactory
     */
    public function __construct(
        CustomerFactory $customerFactory,
        AddressFactory $addressFactory,
        ProductRepositoryInterface $productRepository,
        ScopeConfigInterface $scopeConfig,
        RateRequestFactory $rateRequestFactory,
        MatrixrateFactory $carrierFactory,
        DataObjectFactory $objectFactory,
        OrderExtensionFactory $orderExtensionFactory,
        OrderItemExtensionFactory $orderItemExtensionFactory,
        SerializerInterface $serializer,
        AdditionalDataFactory $additionalDataFactory
    ) {
        $this->customerFactory = $customerFactory;
        $this->addressFactory = $addressFactory;
        $this->productRepository = $productRepository;
        $this->scopeConfig = $scopeConfig;
        $this->rateRequestFactory = $rateRequestFactory;
        $this->carrierFactory = $carrierFactory;
        $this->objectFactory = $objectFactory;
        $this->orderExtensionFactory = $orderExtensionFactory;
        $this->orderItemExtensionFactory = $orderItemExtensionFactory;
        $this->_serializer = $serializer;
        $this->_additionalDataFactory = $additionalDataFactory;
    }

    /**
     * @param OrderManagementInterface $management
     * @param OrderInterface $order
     * @return array
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    public function beforePlace(OrderManagementInterface $management, OrderInterface $order)
    {
        $order_extension_attributes = $order->getExtensionAttributes();
        $order_extension = $order_extension_attributes ?: $this->orderExtensionFactory->create();
        $order_fields_config = $this->scopeConfig->getValue('oad/general/order_data_fields', ScopeInterface::SCOPE_STORE);
        $additional_data = [];
        $source_objects = [];
        if ($order_fields_config === null || empty($order_fields_config)) {
            return [$order];
        }
        $order_fields = $this->_serializer->unserialize($order_fields_config);
        $source_objects['free_shipping'] = $this->simulateShippingAmount($order);
        if ($order_fields) {
            foreach ($order_fields as $order_field) {
                if ($order->getCustomerId() > 0 && $order_field['object_type'] === 'customer' && !array_key_exists('customer', $source_objects)) {
                    $source_objects['customer'] = $this->customerFactory->create()->load($order->getCustomerId());
                }
                if ($order_field['object_type'] === 'billing_address' && !array_key_exists('billing_address', $source_objects)) {
                    $billing_address = $order->getBillingAddress();
                    if ($billing_address->getCustomerAddressId()) {
                        $customer_billing_address = $this->addressFactory->create()->load($billing_address->getCustomerAddressId());
                        $source_objects['billing_address'] = $customer_billing_address;
                    } else {
                        $source_objects['billing_address'] = $billing_address;
                    }
                }
                if ($order_field['object_type'] === 'shipping_address' && !array_key_exists('shipping_address', $source_objects)) {
                    $shipping_address = $order->getShippingAddress();
                    if ($shipping_address->getCustomerAddressId()) {
                        $customer_shipping_address = $this->addressFactory->create()->load($shipping_address->getCustomerAddressId());
                        $source_objects['shipping_address'] = $customer_shipping_address;
                    } else {
                        $source_objects['shipping_address'] = $shipping_address;
                    }
                }
                if (array_key_exists($order_field['object_type'], $source_objects)) {
                    $data_object = $source_objects[$order_field['object_type']];
                    if ($data_object) {
                        $additional_data[$order_field['fieldname']] = $data_object->getData($order_field['object_field']);
                    }
                }
            }
        }
        $additionalDataObject = $this->_additionalDataFactory->create();
        foreach ($additional_data as $field => $value) {
            $additionalDataObject->setCustomAttribute($field, $value);
        }
        $order_extension->setFbOad($additionalDataObject);
        $order->setData('fbnetsuite_base_theoric_shipping_amount', $source_objects['free_shipping']->getValue());
        $order->setData('fbnetsuite_theoric_shipping_amount', $source_objects['free_shipping']->getValue());
        $order->setExtensionAttributes($order_extension);

        $item_fields = $this->scopeConfig->getValue('oad/general/item_data_fields', ScopeInterface::SCOPE_STORE);
        if ($item_fields) {
            $item_fields = $this->_serializer->unserialize($item_fields);
        }
        $items = $order->getAllItems();
        $source_item_objects = ['product' => [], 'category' => []];
        $item_object_ids = [];
        foreach ($items as $key => $order_item) {
            if ($order_item->getProductType() !== 'configurable') {
                $product_id = $order_item->getProductId();
            } else {
                continue;
            }
            if (!array_key_exists("product_{$product_id}", $source_item_objects['product'])) {
                $source_item_objects['product']["product_{$product_id}"] = $this->productRepository->getById($product_id, false, $order->getStoreId(), true);
            }
            $item_object_ids['item_' . $order_item->getQuoteItemId()] = ['product' => $product_id];
            if ($order_item->getParentItem()) {
                $item_object_ids['item_' . $order_item->getParentItem()->getQuoteItemId()] = ['product' => $product_id];
            }
        }
        foreach ($items as $key => $order_item) {
            $additional_data = [];
            $item_extension_attributes = $order_item->getExtensionAttributes();
            $item_extension = $item_extension_attributes ?: $this->orderItemExtensionFactory->create();

            if (array_key_exists('item_' . $order_item->getQuoteItemId(), $item_object_ids)) {
                $product_id = $item_object_ids['item_' . $order_item->getQuoteItemId()]['product'];

                $order_item->setData('kpi_full_price', $source_item_objects['product']["product_{$product_id}"]->getData('price'));
                if ($item_fields) {
                    foreach ($item_fields as $item_field) {
                        if ($item_field['object_type'] === 'product') {
                            $data_object = $source_item_objects['product']["product_{$product_id}"];
                        }
                        $additional_data[$item_field['fieldname']] = $data_object->getData($item_field['object_field']);
                    }
                }
                $itemDataObject[$key] = $this->_additionalDataFactory->create();
                foreach ($additional_data as $field => $value) {
                    $itemDataObject[$key]->setCustomAttribute($field, $value);
                }
                $item_extension->setFbOad($itemDataObject[$key]);
                $order_item->setExtensionAttributes($item_extension);
            }
            $items[$key] = $order_item;
        }
        $order->setItems($items);

        return [$order];
    }

    /**
     * Simulate Shipping Costs in case of Free Shipping.
     *
     * @param OrderInterface $order
     * @return DataObject
     */
    protected function simulateShippingAmount(OrderInterface $order)
    {
        $shipping_amounts = [0];

        // matrix rate
        if ($order->getShippingAmount() == 0 && $order->getShippingAddress()) {
            $address = $order->getShippingAddress();
            $store = $order->getStore();
            $request = $this->rateRequestFactory->create();
            $request->setDestCountryId($address->getCountryId());
            $request->setDestRegionId($address->getRegionId());
            $request->setDestPostcode($address->getPostcode());
            $request->setDestCity($address->getCity());
            $request->setPackageValue($order->getBaseSubtotal());
            $request->setPackageValueWithDiscount($order->getBaseSubtotal());
            $request->setPackageWeight($order->getWeight());
            $request->setFreeMethodWeight($order->getWeight());
            $request->setPackageQty($order->getTotalQtyOrdered());
            $request->setStoreId($store->getId());
            $request->setWebsiteId($store->getWebsiteId());
            $request->setBaseCurrency($order->getBaseCurrencyCode());
            $request->setPackageCurrency($order->getOrderCurrencyCode());
            $request->setLimitCarrier(null);
            $request->setFreeShipping(null);

            //get rates for the matrixrate request created
            $matrixrate = $this->carrierFactory->create();
            //$rates = $matrixrate->getRate($request, false);
            $result = $matrixrate->collectRates($request);
            if ($result) {
                $rates = $result->getAllRates();
                if (!empty($rates)) {
                    foreach ($rates as $rate) {
                        $price = $rate->getPrice();
                        if (isset($price) && $price > 0) {
                            $shipping_amounts[] = $price;
                        }
                    }
                }
            }
        }
        $shippingObject = $this->objectFactory->create();
        $shippingObject->setData('value', max($shipping_amounts));
        return $shippingObject;
    }
}
